from .base import BaseClient


class BybitUnified(BaseClient):
    BASE_ENDPOINT = "https://api.bybit.com"

    def __init__(self):
        super().__init__()
        self.base_endpoint = self.BASE_ENDPOINT

    async def _get_exchange_info(self, category: str) -> dict:
        return await self._get(self.base_endpoint + "/v5/market/instruments-info", params={"category": category})

    async def _get_tickers(self, category: str) -> dict:
        return await self._get(self.base_endpoint + "/v5/market/tickers", params={"category": category})

    async def _get_klines(
        self, symbol: str, interval: str, category: str = None, start: int = None, end: int = None, limit: int = None
    ):
        params = {
            k: v
            for k, v in {
                "category": category,
                "symbol": symbol,
                "interval": interval,
                "start": start,
                "end": end,
                "limit": limit,
            }.items()
            if v is not None
        }

        return await self._get(self.base_endpoint + "/v5/market/kline", params=params)



import pandas as pd


def query_dict(dictionary: dict, query: str, query_env: dict = None) -> dict:
    """
    Query a dictionary with a query string
    :param dictionary: dictionary to query
    :param query: query string
    :param query_env: additional variables for query execution
    :return: queried dictionary
    """
    if not query:
        return dictionary

    df = pd.DataFrame(dictionary).T

    if query_env:
        df = df.query(query, local_dict=query_env)
    else:
        df = df.query(query)

    return df.to_dict(orient="index")


def nested_query_dict(dictionary: dict, key: str, query: str) -> dict:
    """
    Query a dictionary with a query string
    :param dictionary: dictionary to query
    :param query: query string
    :return: queried dictionary
    """
    if not query:
        return dictionary

    new_dict = {outer: inner[key] for outer, inner in dictionary.items() if key in inner}

    queried_dict = query_dict(new_dict, query)
    return {key: dictionary[key] for key in queried_dict.keys()}


def sort_dict(dictionary: dict, ascending: bool = True, num: int = None) -> dict:
    """
    Sort a dictionary by its values
    :param dictionary: dictionary to sort
    :param ascending: ascending or descending
    :param num: number of items to return
    :return: sorted dictionary
    """
    df = pd.DataFrame(dictionary).T.reset_index().sort_values(by="index", ascending=ascending).set_index("index")
    if num:
        df = df.iloc[-num:]
    return df.to_dict(orient="index")



from .base import BaseClient


class HtxUnified(BaseClient):
    BASE_URL = "https://api.huobi.pro"

    def __init__(self):
        super().__init__()
        self.base_endpoint = self.BASE_URL

    async def _get_exchange_info(self):
        return await self._get(self.base_endpoint + "/v2/settings/common/symbols")

    async def _get_tickers(self):
        return await self._get(self.base_endpoint + "/market/tickers")

    async def _get_klines(self, symbol: str, period: str, limit: int):
        params = {"symbol": symbol, "period": period, "size": limit}

        return await self._get(self.base_endpoint + "/market/history/kline", params=params)


class HtxFutures(BaseClient):
    BASE_URL = "https://api.hbdm.com"

    def __init__(self):
        super().__init__()
        self.base_endpoint = self.BASE_URL

    async def _get_linear_contract_info(self):
        return await self._get(self.base_endpoint + "/linear-swap-api/v1/swap_contract_info")

    async def _get_inverse_futures_info(self):
        return await self._get(self.base_endpoint + "/api/v1/contract_contract_info")

    async def _get_inverse_perp_info(self):
        return await self._get(self.base_endpoint + "/swap-api/v1/swap_contract_info")

    async def _get_linear_contract_tickers(self):
        return await self._get(self.base_endpoint + "/v2/linear-swap-ex/market/detail/batch_merged")

    async def _get_inverse_perp_tickers(self):
        return await self._get(self.base_endpoint + "/v2/swap-ex/market/detail/batch_merged")

    async def _get_inverse_futures_tickers(self):
        return await self._get(self.base_endpoint + "/v2/market/detail/batch_merged")

    async def _get_linear_contract_klines(
        self, symbol: str, period: str, start: int = None, end: int = None, limit: int = None
    ):
        params = {"contract_code": symbol, "period": period}
        if start:
            params["from"] = start
        if end:
            params["to"] = end
        if limit:
            params["size"] = limit

        return await self._get(self.base_endpoint + "/linear-swap-ex/market/history/kline", params=params)

    async def _get_inverse_perp_klines(
        self, symbol: str, period: str, start: int = None, end: int = None, limit: int = None
    ):
        params = {"contract_code": symbol, "period": period}
        if start:
            params["from"] = start
        if end:
            params["to"] = end
        if limit:
            params["size"] = limit

        return await self._get(self.base_endpoint + "/swap-ex/market/history/kline", params=params)

    async def _get_inverse_futures_klines(
        self, symbol: str, period: str, start: int = None, end: int = None, limit: int = None
    ):
        params = {"symbol": symbol, "period": period}
        if start:
            params["from"] = start
        if end:
            params["to"] = end
        if limit:
            params["size"] = limit

        return await self._get(self.base_endpoint + "/market/history/kline", params=params)



import tracemalloc

from .exchanges.okx import OkxUnified
from .parsers.okx import OkxParser

tracemalloc.start()


class Okx(OkxUnified):
    name = "okx"
    market_type_map = {"spot": "SPOT", "margin": "MARGIN", "futures": "FUTURES", "perp": "SWAP"}
    _market_type_map = {"SPOT": "spot", "MARGIN": "margin", "FUTURES": "futures", "SWAP": "perp"}

    def __init__(self, api_key: str = None, api_secret: str = None, passphrase: str = None, flag: str = "1"):
        super().__init__(api_key=api_key, api_secret=api_secret, passphrase=passphrase, flag=flag)

        self.parser = OkxParser()
        self.exchange_info = {}

    async def sync_exchange_info(self):
        self.exchange_info = await self.get_exchange_info()

    async def get_exchange_info(self, market_type: str = None):
        if market_type:
            parser = (
                self.parser.spot_margin_exchange_info_parser
                if market_type in ["spot", "margin"]
                else self.parser.futures_perp_exchange_info_parser
            )
            return self.parser.parse_exchange_info(
                await self._get_exchange_info(self.market_type_map[market_type]), parser
            )

        else:
            spot = self.parser.parse_exchange_info(
                await self._get_exchange_info("SPOT"), self.parser.spot_margin_exchange_info_parser
            )
            margin = self.parser.parse_exchange_info(
                await self._get_exchange_info("MARGIN"), self.parser.spot_margin_exchange_info_parser
            )
            futures = self.parser.parse_exchange_info(
                await self._get_exchange_info("FUTURES"), self.parser.futures_perp_exchange_info_parser
            )
            perp = self.parser.parse_exchange_info(
                await self._get_exchange_info("SWAP"), self.parser.futures_perp_exchange_info_parser
            )
            exchange_info = {**self.parser.combine_spot_margin_exchange_info(spot, margin), **futures, **perp}
        return exchange_info

    async def get_tickers(self, market_type: str = None) -> dict:

        if market_type == "spot":
            return self.parser.parse_tickers(await self._get_tickers("SPOT"), "spot", self.exchange_info)
        elif market_type == "futures":
            return self.parser.parse_tickers(await self._get_tickers("FUTURES"), "futures", self.exchange_info)
        elif market_type == "perp":
            return self.parser.parse_tickers(await self._get_tickers("SWAP"), "perp", self.exchange_info)
        else:
            results = {}
            for market_type in ["spot", "futures", "perp"]:
                _market_type = self.market_type_map[market_type]
                parsed_tickers = self.parser.parse_tickers(
                    await self._get_tickers(_market_type), market_type, self.exchange_info
                )
                results.update(parsed_tickers)

            return results

    async def get_ticker(self, instrument_id: str):
        _instrument_id = self.exchange_info[instrument_id]["raw_data"]["instId"]
        market_type = self._market_type_map[self.exchange_info[instrument_id]["raw_data"]["instType"]]
        return {instrument_id: self.parser.parse_ticker(await self._get_ticker(_instrument_id), market_type)}

    async def get_klines(self, instrument_id: str, interval: str, start: int = None, end: int = None, num: int = None):
        info = self.exchange_info[instrument_id]
        market_type = self._market_type_map[info["raw_data"]["instType"]]
        _instrument_id = info["raw_data"]["instId"]
        _interval = self.parser.get_interval(interval)
        limit = 300

        params = {"instId": _instrument_id, "bar": _interval, "limit": limit}

        results = {}
        if start and end:
            query_end = end
            while True:
                params.update({"after": query_end})
                datas = self.parser.parse_klines(
                    await self._get_klines(**params),
                    market_type,
                )
                results.update(datas)

                if not datas or len(datas) < limit:
                    break
                query_end = sorted(datas.keys())[0]
                if query_end < start:
                    break
            results = {k: v for k, v in results.items() if start <= k <= end}

        elif start:
            query_end = end
            while True:
                params.update({"after": query_end} if query_end else {})
                datas = self.parser.parse_klines(
                    await self._get_klines(**params),
                    market_type,
                )
                results.update(datas)
                if not datas or len(datas) < limit:
                    break
                query_end = sorted(datas.keys())[0]
                if query_end < start:
                    break
            results = {k: v for k, v in results.items() if k >= start}
        elif end and num:
            query_end = end
            query_num = min(num, limit)
            while True:
                params.update({"after": query_end, "limit": query_num})
                datas = self.parser.parse_klines(
                    await self._get_klines(**params),
                    market_type,
                )
                results.update(datas)
                if not datas or len(datas) < limit:
                    break
                query_num = min(num - len(results), limit)
                query_end = sorted(datas.keys())[0]
            results = {k: v for k, v in results.items() if k <= end}
        elif num:
            query_end = end
            query_num = min(num, limit)
            while True:
                params.update({"after": query_end, "limit": query_num} if query_end else {"limit": query_num})
                datas = self.parser.parse_klines(
                    await self._get_klines(**params),
                    market_type,
                )
                results.update(datas)
                if not datas or len(datas) < limit:
                    break
                query_num = min(num - len(results), limit)
                query_end = sorted(datas.keys())[0]

            results = dict(sorted(results.items(), key=lambda x: x[0])[-num:])
        else:
            raise Exception("invalid params")

        return results

    async def get_balance(self):
        return self.parser.parse_balance(await self._get_balance())

    async def get_positions(self):
        return self.parser.parse_positions(await self._get_positions(), self.exchange_info)

    async def get_account_info(self):
        return self.parser.parse_account_config(await self._get_account_config())

    async def place_market_order(self, instrument_id: str, side: str, volume: float, in_quote: bool = False):
        if instrument_id not in self.exchange_info:
            raise Exception(f"{instrument_id} not found in exchange_info")

        info = self.exchange_info[instrument_id]
        _instrument_id = info["raw_data"]["instId"]
        _order_type = "market"

        order_id = self.parser.parse_order_id(
            await self._place_order(
                instId=_instrument_id,
                side=side,
                sz=str(volume),
                ordType=_order_type,
                tgtCcy="quote_ccy" if in_quote else "base_ccy",
            )
        )
        return self.parser.parse_order_info(await self._get_order_info(_instrument_id, str(order_id)), info)

    async def place_limit_order(
        self, instrument_id: str, side: str, price: float, volume: float, in_quote: bool = False
    ):
        if instrument_id not in self.exchange_info:
            raise Exception(f"{instrument_id} not found in exchange_info")

        info = self.exchange_info[instrument_id]
        _instrument_id = info["raw_data"]["instId"]
        _order_type = "limit"

        order_id = self.parser.parse_order_id(
            await self._place_order(
                instId=_instrument_id,
                side=side,
                sz=str(volume),
                px=str(price),
                ordType=_order_type,
                tgtCcy="quote_ccy" if in_quote else "base_ccy",
            )
        )
        return self.parser.parse_order_info(await self._get_order_info(_instrument_id, str(order_id)), info)

    async def cancel_order(self, instrument_id: str, order_id: int):
        if instrument_id not in self.exchange_info:
            raise Exception(f"{instrument_id} not found in exchange_info")
        info = self.exchange_info[instrument_id]
        _instrument_id = info["raw_data"]["instId"]
        return self.parser.parse_cancel_order(await self._cancel_order(_instrument_id, str(order_id)))

    async def get_opened_orders(self, market_type: str = None, instrument_id: str = None) -> list:
        results = []
        params = {"limit": "100"}
        if market_type:
            _market_type = self.market_type_map[market_type]
            params["instType"] = _market_type
            results = self.parser.parse_opened_orders(
                await self._get_opended_orders(**params), infos=self.exchange_info
            )
        elif instrument_id:
            if instrument_id not in self.exchange_info:
                raise Exception(f"{instrument_id} not found in exchange_info")
            info = self.exchange_info[instrument_id]
            _instrument_id = info["raw_data"]["instId"]
            params["instId"] = _instrument_id
            results = self.parser.parse_opened_orders(await self._get_opended_orders(**params), info=info)
        else:
            raise Exception("market_type or instrument_id must be provided")

        return results

    async def get_history_orders(self, market_type: str = None, instrument_id: str = None) -> list:
        results = []
        params = {"limit": "100"}

        if market_type:
            pass
        elif instrument_id:
            if instrument_id not in self.exchange_info:
                raise Exception
            info = self.exchange_info[instrument_id]
            _instrument_id = info["raw_data"]["instId"]
            _market = info["raw_data"]["instType"]
            params.update({"instId": _instrument_id, "instType": _market})
            results = self.parser.parse_history_orders(
                await self._get_history_orders(**params), infos=self.exchange_info
            )

        else:
            raise Exception("market_type or instrument_id must be provided")

        return results

    async def get_history_funding_rate(
        self, instrument_id: str, start: int = None, end: int = None, num: int = 30
    ) -> list:
        if instrument_id not in self.exchange_info:
            raise Exception(f"{instrument_id} not found in exchange_info")

        info = self.exchange_info[instrument_id]
        _instrument_id = info["raw_data"]["instId"]
        limit = 100
        params = {"instId": _instrument_id, "limit": limit}
        results = []
        query_end = None

        if start and end:
            query_end = end + 1
            while True:
                params.update({"after": query_end})
                result = self.parser.parse_funding_rates(
                    await self._get_history_funding_rate(**params),
                    info,
                )
                results.extend(result)
                if not result or len(result) < limit:
                    break
                query_end = min([v["timestamp"] for v in result])

                if query_end < start:
                    break
                continue
            return [v for v in results if start <= v["timestamp"] <= end]

        elif num:
            while True:
                params.update({"after": query_end} if query_end else {})
                result = self.parser.parse_funding_rates(
                    await self._get_history_funding_rate(**params),
                    info,
                )
                results.extend(result)

                # exclude same timestamp
                results = list({v["timestamp"]: v for v in results}.values())

                if not result or len(result) < limit or len(results) >= num:
                    break

                query_end = min([v["timestamp"] for v in result])
                continue
            return sorted(results, key=lambda x: x["timestamp"], reverse=False)[-num:]

        else:
            raise Exception("(start, end) or num must be provided")

    async def get_current_funding_rate(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise Exception(f"{instrument_id} not found in exchange_info")
        info = self.exchange_info[instrument_id]
        _instrument_id = info["raw_data"]["instId"]
        return self.parser.parse_current_funding_rate(await self._get_current_funding_rate(_instrument_id), info)



from datetime import timedelta as td

from .base import Parser


class OkxParser(Parser):
    interval_map = {
        "1m": "1m",
        "3m": "3m",
        "5m": "5m",
        "15m": "15m",
        "30m": "30m",
        "1h": "1H",
        "2h": "2H",
        "4h": "4H",
        "6h": "6H",
        "12h": "12H",
        "1d": "1D",
        "2d": "2D",
        "3d": "3D",
        "1w": "1W",
        "1M": "1M",
        "3M": "3M",
    }

    market_type_map = {"spot": "SPOT", "margin": "MARGIN", "futures": "FUTURES", "perp": "SWAP"}
    _market_type_map = {"SPOT": "spot", "MARGIN": "margin", "FUTURES": "futures", "SWAP": "perp"}

    @staticmethod
    def check_response(response: dict):
        if response.get("code") == "0":
            return {"code": 200, "status": "success", "data": response["data"]}
        else:
            raise Exception(f"Error {response['code']} {response['msg']}")

    def _parse_leverage(self, lever: any):
        return int(lever) if lever else 1

    def _parse_symbol(self, data: dict) -> str:
        market_type = data["instType"]
        if market_type in ["SPOT", "MARGIN"]:
            symbol = f"{data['baseCcy']}/{data['quoteCcy']}"
        elif market_type in ["SWAP", "FUTURES"]:
            if data["ctType"] == "linear":
                symbol = f"{data['ctValCcy']}/{data['settleCcy']}"
            else:
                symbol = f"{data['settleCcy']}/{data['ctValCcy']}"
        return symbol

    @property
    def spot_margin_exchange_info_parser(self):
        return {
            "active": (lambda x: x["state"] == "live"),
            "is_spot": (lambda x: self.parse_is_spot(x["instType"])),
            "is_margin": (lambda x: self.parse_is_margin(x["instType"])),
            "is_futures": False,
            "is_perp": False,
            "is_linear": True,
            "is_inverse": False,
            "symbol": (lambda x: self._parse_symbol(x)),
            "base": (lambda x: self.parse_base_currency(x["baseCcy"])),
            "quote": (lambda x: str(x["quoteCcy"])),
            "settle": (lambda x: str(x["quoteCcy"])),
            "multiplier": 1,  # spot and margin default multiplier is 1
            "leverage": (lambda x: self._parse_leverage(x["lever"])),
            "listing_time": (lambda x: int(x["listTime"])),
            "expiration_time": None,  # spot not support this field
            "contract_size": 1,
            "tick_size": (lambda x: float(x["tickSz"])),
            "min_order_size": (lambda x: float(x["minSz"])),
            "max_order_size": (lambda x: float(x["maxMktSz"])),
            "raw_data": (lambda x: x),
        }

    @property
    def futures_perp_exchange_info_parser(self):
        return {
            "active": (lambda x: x["state"] == "live"),
            "is_spot": False,
            "is_margin": False,
            "is_futures": (lambda x: self.parse_is_futures(x["instType"])),
            "is_perp": (lambda x: self.parse_is_perpetual(x["instType"])),
            "is_linear": (lambda x: self.parse_is_linear(x["ctType"])),
            "is_inverse": (lambda x: self.parse_is_inverse(x["ctType"])),
            "symbol": (lambda x: self._parse_symbol(x)),
            "base": (lambda x: self.parse_base_currency(x["ctValCcy"] if x["ctType"] == "linear" else x["settleCcy"])),
            "quote": (lambda x: str(x["settleCcy"] if x["ctType"] == "linear" else x["ctValCcy"])),
            "settle": (lambda x: str(x["settleCcy"])),
            "multiplier": (lambda x: self.parse_multiplier(x["ctMult"])),
            "leverage": (lambda x: self._parse_leverage(x["lever"])),
            "listing_time": (lambda x: int(x["listTime"])),
            "expiration_time": (lambda x: int(x["expTime"]) if x["expTime"] else None),
            "contract_size": (
                lambda x: float(x["ctVal"])
            ),  # linear will be how many base ccy and inverse will be how many quote ccy
            "tick_size": (lambda x: float(x["tickSz"])),
            "min_order_size": (lambda x: float(x["minSz"])),
            "max_order_size": (lambda x: float(x["maxMktSz"])),
            "raw_data": (lambda x: x),
        }

    def parse_exchange_info(self, response: dict, parser: dict) -> dict:
        response = self.check_response(response)

        datas = response["data"]
        results = {}
        for data in datas:
            result = self.get_result_with_parser(data, parser)
            id = self.parse_unified_id(result)
            results[id] = result
        return results

    def combine_spot_margin_exchange_info(self, spots: dict, margins: dict) -> dict:
        for instrument_id in margins.keys():
            margin = margins[instrument_id]
            if instrument_id in spots.keys() and margin["active"]:
                spot = spots[instrument_id]

                spot["is_margin"] = True
                spot["leverage"] = margin["leverage"]

                if spot["min_order_size"] != margin["min_order_size"]:
                    print(f"min_order_size of {instrument_id} is different between spot and margin")

                if spot["max_order_size"] != margin["max_order_size"]:
                    print(f"max_order_size of {instrument_id} is different between spot and margin")

                if spot["contract_size"] != margin["contract_size"]:
                    print(f"contract_size of {instrument_id} is different between spot and margin")

                if spot["tick_size"] != margin["tick_size"]:
                    print(f"tick_size of {instrument_id} is different between spot and margin")

                spots[instrument_id] = spot
        return spots

    def parse_ticker(self, response: any, market_type: str) -> dict:
        if "data" in response:
            response = response["data"][0]

        if market_type == "spot":
            base_volume = float(response["vol24h"])
            quote_volume = float(response["volCcy24h"])
        else:
            base_volume = float(response["volCcy24h"])
            quote_volume = float(response["volCcy24h"]) * (float(response["last"]) + float(response["open24h"])) / 2

        return {
            "symbol": response["instId"],
            "open_time": self.adjust_timestamp(int(response["ts"]), td(days=-1)),
            "close_time": int(response["ts"]),
            "open": float(response["open24h"]),
            "high": float(response["high24h"]),
            "low": float(response["low24h"]),
            "last_price": float(response["last"]),
            "base_volume": base_volume,
            "quote_volume": quote_volume,
            "price_change": None,
            "price_change_percent": None,
            "raw_data": response,
        }

    def get_id_map(self, infos: dict, market_type: str = None) -> dict:
        if market_type:
            infos = self.query_dict(infos, {f"is_{market_type}": True})
        return {v["raw_data"]["instId"]: k for k, v in infos.items()}

    def parse_tickers(self, response: dict, market_type: str, infos: dict) -> dict:
        response = self.check_response(response)
        datas = response["data"]

        id_map = self.get_id_map(infos, market_type)
        results = {}
        for data in datas:
            instrument_id = id_map[data["instId"]]
            results[instrument_id] = self.parse_ticker(data, market_type)
        return results

    @staticmethod
    def parse_kline(response: dict, market_type: str) -> dict:
        return {
            "open": float(response[1]),
            "high": float(response[2]),
            "low": float(response[3]),
            "close": float(response[4]),
            "base_volume": float(response[6] if market_type != "spot" else response[5]),
            "quote_volume": float(response[7]),
            "raw_data": response,
        }

    def parse_klines(self, response: dict, market_type: str) -> dict:
        response = self.check_response(response)

        datas = response["data"]
        results = {}
        for data in datas:
            result = self.parse_kline(data, market_type)
            timestamp = int(data[0])
            results[timestamp] = result
        return results

    def parse_funding_rates(self, response: dict, info: dict) -> list:
        response = self.check_response(response)
        datas = response["data"]

        results = []
        for data in datas:
            results.append(
                {
                    "timestamp": int(data["fundingTime"]),
                    "instrument_id": self.parse_unified_id(info),
                    "market": self._market_type_map[data["instType"]],
                    "funding_rate": float(data["fundingRate"]),
                    "realized_rate": float(data["realizedRate"]),
                    "raw_data": data,
                }
            )
        return results

    def parse_current_funding_rate(self, resposne: dict, info: dict) -> dict:
        response = self.check_response(resposne)
        data = response["data"][0]
        return {
            "timestamp": int(data["ts"]),
            "instrument_id": self.parse_unified_id(info),
            "market": self._market_type_map[data["instType"]],
            "funding_time": int(data["fundingTime"]),
            "funding_rate": float(data["fundingRate"]),
            "next_funding_time": int(data["nextFundingTime"]),
            "next_funding_rate": float(data["nextFundingRate"]) if data["nextFundingRate"] else None,
            "raw_data": data,
        }

    def parse_balance(self, response: dict) -> dict:
        response = self.check_response(response)

        datas = response["data"][0]["details"]
        results = {}
        for data in datas:
            currency = data["ccy"]
            results[currency] = {
                "balance": float(data["cashBal"]),
                "available_balance": float(data["availBal"]),
                "raw_data": data,
            }
        return results

    def parse_positions(self, response: dict, infos: dict) -> dict:
        response = self.check_response(response)
        id_map = self.get_id_map(infos)
        datas = response["data"]

        results = {}
        for data in datas:
            instrument_id = id_map[data["instId"]]
            results[instrument_id] = {"position": float(data["pos"]), "raw_data": data}
        return results

    def parse_account_config(self, response: dict) -> dict:
        response = self.check_response(response)
        data = response["data"][0]

        return {
            "account_id": data["uid"],
            "main_account_id": data["mainUid"],
            "key_name": data["label"],
            "permission": data["perm"].split(","),
            "raw_data": data,
        }

    def get_interval(self, interval: str) -> str:
        return self.interval_map[interval]

    def parse_order_id(self, response: dict) -> int:
        response = self.check_response(response)
        data = response["data"][0]
        return int(data["ordId"])

    def parse_order_info(self, response: dict, info: dict) -> dict:
        response = self.check_response(response)
        data = response["data"][0]
        return {
            "timestamp": int(data["cTime"]),
            "instrument_id": self.parse_unified_id(info),
            "side": data["side"],
            "price": self.parse_str(data["px"], float),
            "volume": self.parse_str(data["sz"], float),
            "fee_ccy": data["feeCcy"],
            "fee": self.parse_str(data["fee"], float),
            "order_id": int(data["ordId"]),
            "order_type": data["ordType"],
            "status": data["state"],
            "raw_data": data,
        }

    def parse_cancel_order(self, response: dict) -> dict:
        response = self.check_response(response)
        data = response["data"][0]
        return {
            "order_id": int(data["ordId"]),
            "raw_data": data,
        }

    def parse_opened_orders(self, response: dict, infos: dict) -> list:
        response = self.check_response(response)
        datas = response["data"]

        results = []

        id_map = self.get_id_map(infos)

        for data in datas:
            results.append(
                {
                    "timestamp": int(data["cTime"]),
                    "instrument_id": id_map[data["instId"]],
                    "market_type": self._market_type_map[data["instType"]],
                    "side": data["side"],
                    "price": self.parse_str(data["px"], float),
                    "volume": self.parse_str(data["sz"], float),
                    "order_id": int(data["ordId"]),
                    "order_type": data["ordType"],
                    "status": data["state"],
                    "raw_data": data,
                }
            )
        return results

    def parse_history_orders(self, response: dict, infos: dict) -> list:
        response = self.check_response(response)
        datas = response["data"]

        id_map = self.get_id_map(infos)

        results = []
        for data in datas:
            results.append(
                {
                    "timestamp": int(data["fillTime"]),
                    "instrument_id": id_map[data["instId"]],
                    "market_type": self._market_type_map[data["instType"]],
                    "side": data["side"],
                    "executed_price": self.parse_str(data["fillPx"], float),
                    "executed_volume": self.parse_str(data["fillSz"], float),
                    "target_price": self.parse_str(data["px"], float),
                    "target_volume": self.parse_str(data["sz"], float),
                    "fee_currency": data["feeCcy"],
                    "fee": self.parse_str(data["fee"], float),
                    "order_type": data["ordType"],
                    "order_id": int(data["ordId"]),
                    "status": data["state"],
                    "raw_data": data,
                }
            )
        return results



from .exchanges.htx import HtxFutures, HtxUnified
from .parsers.htx import HtxParser
from .utils import query_dict, sort_dict


class Htx(object):
    name = "htx"

    def __init__(self):
        self.spot = HtxUnified()
        self.futures = HtxFutures()
        self.parser = HtxParser()
        self.exchange_info = {}

    async def close(self):
        await self.spot.close()
        await self.futures.close()

    @classmethod
    async def create(cls):
        instance = cls()
        instance.exchange_info = await instance.get_exchange_info()
        return instance

    async def get_exchange_info(self, market_type: str = None):
        spot = self.parser.parse_exchange_info(
            await self.spot._get_exchange_info(), self.parser.spot_exchange_info_parser
        )
        linear = self.parser.parse_exchange_info(
            await self.futures._get_linear_contract_info(), self.parser.linear_exchange_info_parser
        )
        inverse_futures = self.parser.parse_exchange_info(
            await self.futures._get_inverse_futures_info(), self.parser.inverse_futures_exchange_info_parser
        )
        inverse_perp = self.parser.parse_exchange_info(
            await self.futures._get_inverse_perp_info(), self.parser.inverse_perp_exchange_info_parser
        )

        return {**spot, **linear, **inverse_futures, **inverse_perp}

    async def get_tickers(self, market_type: str = None):
        # get_all tickers then filter by market_type
        if market_type:
            if market_type == "spot":
                return self.parser.parse_tickers(await self.spot._get_tickers(), self.exchange_info, "spot")
            else:
                linear = self.parser.parse_tickers(
                    await self.futures._get_linear_contract_tickers(), self.exchange_info, "linear"
                )
                inverse_perp = self.parser.parse_tickers(
                    await self.futures._get_inverse_perp_tickers(), self.exchange_info, "inverse_perp"
                )

                inverse_futures = self.parser.parse_tickers(
                    await self.futures._get_inverse_futures_tickers(), self.exchange_info, "inverse_futures"
                )
                results = {**linear, **inverse_perp, **inverse_futures}

                instrument_id = list(query_dict(self.exchange_info, f"is_{market_type} == True").keys())
                return {k: v for k, v in results.items() if k in instrument_id}
        else:
            spot = self.parser.parse_tickers(await self.spot._get_tickers(), self.exchange_info, "spot")

            linear = self.parser.parse_tickers(
                await self.futures._get_linear_contract_tickers(), self.exchange_info, "linear"
            )
            inverse_perp = self.parser.parse_tickers(
                await self.futures._get_inverse_perp_tickers(), self.exchange_info, "inverse_perp"
            )
            inverse_futures = self.parser.parse_tickers(
                await self.futures._get_inverse_futures_tickers(), self.exchange_info, "inverse_futures"
            )
            return {**spot, **linear, **inverse_perp, **inverse_futures}

    async def get_klines(self, instrument_id: str, interval: str, start: int = None, end: int = None, num: int = None):
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not in exchange_info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)
        _interval = self.parser.get_interval(interval, market_type)

        symbol_map = {
            "spot": "sc",
            "linear": "contract_code",
            "inverse_futures": "symbol",
            "inverse_perp": "contract_code",
        }

        method_map = {
            "spot": self.spot._get_klines,
            "linear": self.futures._get_linear_contract_klines,
            "inverse_futures": self.futures._get_inverse_futures_klines,
            "inverse_perp": self.futures._get_inverse_perp_klines,
        }

        limit_map = {
            "spot": 2000,
            "linear": 2000,
            "inverse_futures": 2000,
            "inverse_perp": 2000,
        }

        params = {
            "symbol": info["raw_data"][symbol_map[market_type]]
            if market_type != "inverse_futures"
            else self.parser.parse_inverse_futures_symbol(info["raw_data"]),
            "period": _interval,
            "limit": limit_map[market_type],
        }
        results = {}
        query_end = None
        if start and end and market_type != "spot":
            query_end = end
            while True:
                params["end"] = str(query_end)[:10]
                result = self.parser.parse_klines(await method_map[market_type](**params), market_type, info)
                results.update(result)

                temp_end = str(sorted(list(result.keys()))[0])[:10]
                if len(result) < limit_map[market_type] or temp_end == query_end:
                    break

                query_end = temp_end  # get the earliest timestamp in 10 digits
                if query_end <= start:
                    break
            return {k: v for k, v in results.items() if start <= int(k) <= end}

        elif num:
            while True:
                params.update({"end": query_end} if query_end else {})
                result = self.parser.parse_klines(await method_map[market_type](**params), market_type, info)
                results.update(result)

                temp_end = str(sorted(list(result.keys()))[0])[:10]
                if len(result) < limit_map[market_type] or temp_end == query_end:
                    break

                query_end = temp_end  # get the earliest timestamp in 10 digits
                if len(results) >= num:
                    break
            return sort_dict(results, ascending=True, num=num)

        else:
            raise ValueError("(start, end) or num must be provided")



import tracemalloc
from typing import Literal, Optional

from .exchanges.binance import BinanceInverse, BinanceLinear, BinanceSpot
from .parsers.binance import BinanceParser
from .utils import sort_dict

tracemalloc.start()


class Binance(object):
    name = "binance"

    def __init__(self, api_key: str = None, api_secret: str = None):
        self.spot = BinanceSpot(api_key=api_key, api_secret=api_secret)
        self.linear = BinanceLinear()
        self.inverse = BinanceInverse()
        self.parser = BinanceParser()

        self.exchange_info = {}

    async def close(self):
        await self.spot.close()
        await self.linear.close()
        await self.inverse.close()

    @classmethod
    async def create(cls):
        instance = cls()
        instance.exchange_info = await instance.get_exchange_info()
        return instance

    async def sync_exchange_info(self) -> None:
        self.exchange_info = await self.get_exchange_info()

    async def get_exchange_info(self, market_type: Optional[Literal["spot", "margin", "futures", "perp"]] = None):
        spot = self.parser.parse_exchange_info(
            await self.spot._get_exchange_info(), self.parser.spot_exchange_info_parser
        )
        linear = self.parser.parse_exchange_info(
            await self.linear._get_exchange_info(), self.parser.futures_exchange_info_parser("linear")
        )
        inverse = self.parser.parse_exchange_info(
            await self.inverse._get_exchange_info(), self.parser.futures_exchange_info_parser("inverse")
        )
        result = {**spot, **linear, **inverse}
        return result if not market_type else self.parser.query_dict(result, {f"is_{market_type}": True})

    async def get_ticker(self, instrument_id: str):
        _symbol = self.exchange_info[instrument_id]["raw_data"]["symbol"]
        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)

        if market_type == "spot":
            return {instrument_id: self.parser.parse_ticker(await self.spot._get_ticker(_symbol), info)}
        elif market_type == "linear":
            return {instrument_id: self.parser.parse_ticker(await self.linear._get_ticker(_symbol), info)}
        elif market_type == "inverse":
            return {instrument_id: self.parser.parse_ticker(await self.inverse._get_ticker(_symbol), info)}

    async def get_tickers(self, market_type: Optional[Literal["spot", "margin", "futures", "perp"]] = None) -> dict:
        results = {}

        tickers = [(self.spot, "spot"), (self.linear, "linear"), (self.inverse, "inverse")]

        for exchange, _market_type in tickers:
            parsed_tickers = self.parser.parse_tickers(await exchange._get_tickers(), _market_type, self.exchange_info)
            results.update(parsed_tickers)

        if market_type:
            ids = list(self.parser.query_dict(self.exchange_info, {f"is_{market_type}": True}).keys())
            return self.parser.query_dict_by_keys(results, ids)
        else:
            return results

    async def get_klines(self, instrument_id: str, interval: str, start: int = None, end: int = None, num: int = 500):
        _symbol = self.exchange_info[instrument_id]["raw_data"]["symbol"]
        _interval = self.parser.get_interval(interval)
        limit = 1000
        market_type = self.parser.get_market_type(self.exchange_info[instrument_id])

        params = {
            "symbol": _symbol,
            "interval": _interval,
            "limit": limit,
        }

        method_map = {
            "spot": self.spot._get_klines,
            "linear": self.linear._get_klines,
            "inverse": self.inverse._get_klines,
        }

        query_end = None

        results = {}
        if start and end:
            query_end = end
            while True:
                params["endTime"] = query_end
                klines = self.parser.parse_klines(await method_map[market_type](**params), market_type)
                if not klines:
                    break

                results.update(klines)
                query_end = sorted(list(klines.keys()))[0]
                if len(klines) < limit or query_end <= start:
                    break
                continue
            return sort_dict({k: v for k, v in results.items() if end >= k >= start}, ascending=True)

        elif num:
            while True:
                params.update({"endTime": query_end} if query_end else {})
                klines = self.parser.parse_klines(
                    await method_map[market_type](**params),
                    market_type,
                )
                results.update(klines)
                if len(klines) < limit or len(results) >= num:
                    break
                query_end = sorted(list(klines.keys()))[0]
                continue

            return sort_dict(results, ascending=True, num=num)

    # Private function
    async def get_spot_account_info(self) -> dict:
        return self.parser.parse_spot_account_info(await self.spot._get_account_info())



from datetime import datetime as dt
from datetime import timedelta as td

from .base import Parser


class BybitParser(Parser):
    INTERVAL_MAP = {
        "1m": "1",
        "3m": "3",
        "5m": "5",
        "15m": "15",
        "30m": "30",
        "1h": "60",
        "2h": "120",
        "4h": "240",
        "6h": "360",
        "12h": "720",
        "1d": "D",
        "1M": "M",
        "1w": "W",
    }

    @staticmethod
    def check_response(response: dict):
        if response["retCode"] != 0:
            return {
                "code": 400,
                "status": "error",
                "data": response,
            }
        else:
            return {
                "code": 200,
                "status": "success",
                "data": response["result"]["list"],
            }

    @property
    def spot_exchange_info_parser(self) -> dict:
        return {
            "active": (lambda x: x["status"] == "Trading"),
            "is_spot": True,
            "is_margin": (lambda x: self.parse_is_margin(x["marginTrading"])),
            "is_futures": False,
            "is_perp": False,
            "is_linear": True,
            "is_inverse": False,
            "symbol": (lambda x: self.parse_unified_symbol(x["baseCoin"], x["quoteCoin"])),
            "base": (lambda x: str(x["baseCoin"])),
            "quote": (lambda x: str(x["quoteCoin"])),
            "settle": (lambda x: str(x["quoteCoin"])),
            "multiplier": 1,  # spot multiplier default 1
            "leverage": 1,  # spot leverage default 1
            "listing_time": None,  # api not support this field
            "expiration_time": None,  # spot not support this field
            "contract_size": 1,  # spot contract size default 1
            "tick_size": (lambda x: float(x["priceFilter"]["tickSize"])),
            "min_order_size": (lambda x: float(x["lotSizeFilter"]["minOrderQty"])),
            "max_order_size": (lambda x: float(x["lotSizeFilter"]["maxOrderQty"])),
            "raw_data": (lambda x: x),
        }

    @property
    def perp_futures_exchange_info_parser(self) -> dict:
        return {
            "active": (lambda x: x["status"] == "Trading"),
            "is_spot": False,
            "is_margin": False,
            "is_futures": (lambda x: self.parse_is_futures(x["contractType"])),
            "is_perp": (lambda x: self.parse_is_perpetual(x["contractType"])),
            "is_linear": (lambda x: self.parse_is_linear(x["contractType"])),
            "is_inverse": (lambda x: self.parse_is_inverse(x["contractType"])),
            "symbol": (lambda x: self.parse_unified_symbol(self.parse_base_currency(x["baseCoin"]), x["quoteCoin"])),
            "base": (lambda x: self.parse_base_currency(x["baseCoin"])),
            "quote": (lambda x: str(x["quoteCoin"])),
            "settle": (lambda x: str(x["settleCoin"])),
            "multiplier": (lambda x: self.parse_multiplier(x["baseCoin"])),
            "leverage": (lambda x: float(x["leverageFilter"]["maxLeverage"])),
            "listing_time": (lambda x: int(x["launchTime"])),
            "expiration_time": (lambda x: int(x["deliveryTime"])),
            "contract_size": (lambda x: float(x["lotSizeFilter"]["qtyStep"])),
            "tick_size": (lambda x: float(x["priceFilter"]["tickSize"])),
            "min_order_size": (lambda x: float(x["lotSizeFilter"]["minOrderQty"])),
            "max_order_size": (lambda x: float(x["lotSizeFilter"]["maxOrderQty"])),
            "raw_data": (lambda x: x),
        }

    def parse_exchange_info(self, response: dict, parser: dict) -> dict:
        response = self.check_response(response)
        datas = response["data"]

        results = {}
        for data in datas:
            result = self.get_result_with_parser(data, parser)
            id = self.parse_unified_id(result)
            results[id] = result
        return results

    def parse_tickers(self, response: dict, market_type: str) -> list:
        response = self.check_response(response)
        datas = response["data"]

        results = []
        for data in datas:
            result = self.parse_ticker(data, market_type)
            results.append(result)
        return results

    def parse_ticker(self, response: dict, market_type: str) -> dict:
        return {
            "symbol": response["symbol"],
            "open_time": int((dt.now() - td(days=1)).timestamp() * 1000),
            "close_time": int(dt.now().timestamp() * 1000),
            "open": float(response["prevPrice24h"]),
            "high": float(response["highPrice24h"]),
            "low": float(response["lowPrice24h"]),
            "last_price": float(response["lastPrice"]),
            "base_volume": float(response["volume24h"]) if market_type != "inverse" else float(response["turnover24h"]),
            "quote_volume": float(response["turnover24h"])
            if market_type != "inverse"
            else float(response["volume24h"]),
            "price_change": float(response["prevPrice24h"]) - float(response["lastPrice"]),
            "price_change_percent": float(response["price24hPcnt"]),
            "raw_data": response,
        }

    def get_interval(self, interval: str) -> str:
        if interval not in self.INTERVAL_MAP:
            raise ValueError(f"Invalid interval: {interval}")
        return self.INTERVAL_MAP[interval]

    def parse_klines(self, response: dict) -> dict:
        response = self.check_response(response)
        if response["code"] != 200:
            return response

        results = {}
        datas = response["data"]
        for data in datas:
            result = self.parse_kline(data)
            timestamp = int(data[0])
            results[timestamp] = result
        return results

    def get_category(self, info: dict) -> str:
        if info["is_spot"] or info["is_margin"]:
            return "spot"
        elif info["is_linear"]:
            return "linear"
        elif info["is_inverse"]:
            return "inverse"
        else:
            raise ValueError(f"Invalid market type: {info}")

    def parse_kline(self, response: dict) -> dict:
        return {
            "open": float(response[1]),
            "high": float(response[2]),
            "low": float(response[3]),
            "close": float(response[4]),
            "base_volume": float(response[5]),
            "quote_volume": float(response[6]),
            "raw_data": response,
        }



import base64
import hashlib
import hmac
import json
import time
from datetime import datetime as dt


class OkxAuth(object):
    BASE_ENDPOINT = "https://www.okx.com"

    def __init__(self, api_key, api_secret, passphrase, use_server_time: bool, debug: bool, flag: str):
        self.api_key = api_key
        self.api_secret = api_secret
        self.passphrase = passphrase
        self.use_server_time = use_server_time
        self.debug = debug
        self.flag = flag

        self.body = ""

    def get_private_header(self, method: str, request_path: str, params: dict):
        if method == "GET":
            request_path = (request_path + self.parse_params_to_str(params)).replace(self.BASE_ENDPOINT, "")
        elif method == "POST":
            request_path = request_path.replace(self.BASE_ENDPOINT, "")
        else:
            raise ValueError(f"Invalid method: {method}")

        body = json.dumps(params, separators=(",", ":")) if method == "POST" else ""
        self.body = body
        timestamp = self.get_timestamp()

        sign = self.sign(self.pre_hash(timestamp, method, request_path, body))

        header = dict()
        header["Content-Type"] = "application/json"
        header["OK-ACCESS-KEY"] = self.api_key
        header["OK-ACCESS-SIGN"] = sign.decode("utf-8")
        header["OK-ACCESS-TIMESTAMP"] = str(timestamp)
        header["OK-ACCESS-PASSPHRASE"] = self.passphrase
        header["x-simulated-trading"] = self.flag

        if self.debug:
            print("header: ", header)
        return header

    def sign(self, message):
        mac = hmac.new(bytes(self.api_secret, encoding="utf8"), bytes(message, encoding="utf-8"), digestmod="sha256")
        d = mac.digest()
        return base64.b64encode(d)

    def pre_hash(self, timestamp, method, request_path, body):
        if self.debug:
            print("body: ", body)
        return str(timestamp) + str.upper(method) + request_path + body

    @staticmethod
    def get_timestamp():
        now = dt.utcnow()
        t = now.isoformat("T", "milliseconds")
        return t + "Z"

    @staticmethod
    def parse_params_to_str(params):
        url = "?"
        for key, value in params.items():
            if value != "":
                url = url + str(key) + "=" + str(value) + "&"
        url = url[0:-1]
        return url


class BinanceAuth(object):
    def __init__(self, api_key: str, api_secret: str):
        self.api_key = api_key
        self.api_secret = api_secret

    def get_private_header(self):
        headers = {"X-MBX-APIKEY": self.api_key}
        return headers

    def update_params(self, params: dict):
        params["timestamp"] = int(time.time() * 1000)

        payload = "&".join([f"{k}={v}" for k, v in params.items()])
        signature = hmac.new(self.api_secret.encode("utf-8"), payload.encode("utf-8"), hashlib.sha256).hexdigest()
        params["signature"] = signature
        return params



from .exchanges.gateio import GateioClient
from .parsers.gateio import GateioParser
from .utils import sort_dict


class Gateio(object):
    name = "gateio"

    PERP_SETTLE = ["btc", "usdt", "usd"]

    def __init__(self):
        self.exchange = GateioClient()
        self.parser = GateioParser()
        self.exchange_info = {}

    async def close(self):
        await self.exchange.close()

    @classmethod
    async def create(cls):
        instance = cls()
        instance.exchange_info = await instance.get_exchange_info()
        return instance

    async def get_exchange_info(self):
        spot = self.parser.parse_exchange_info(
            await self.exchange._get_currency_pairs(), self.parser.spot_exchange_info_parser
        )

        perps = {}
        for settle in self.PERP_SETTLE:
            perp = self.parser.parse_exchange_info(
                await self.exchange._get_perp_info(settle),
                self.parser.perp_exchange_info_parser,
                settle=settle.upper(),
            )
            perps.update(perp)

        settle = "usdt"
        futures = self.parser.parse_exchange_info(
            await self.exchange._get_futures_info(settle),
            self.parser.futures_exchange_info_parser,
            settle=settle.upper(),
        )

        return {**spot, **perps, **futures}

    async def get_tickers(self, market_type: str = None) -> dict:
        if market_type == "spot":
            return self.parser.parse_tickers(await self.exchange._get_spot_tickers(), self.exchange_info, "spot")
        elif market_type == "futures":
            return self.parser.parse_tickers(
                await self.exchange._get_futures_tickers(settle="usdt"), self.exchange_info, "futures"
            )
        elif market_type == "perp":
            perps = {}
            for settle in self.PERP_SETTLE:
                perp = self.parser.parse_tickers(
                    await self.exchange._get_perp_tickers(settle), self.exchange_info, "perp"
                )
                perps.update(perp)
            return perps
        else:
            spot = self.parser.parse_tickers(await self.exchange._get_spot_tickers(), self.exchange_info, "spot")
            futures = self.parser.parse_tickers(
                await self.exchange._get_futures_tickers(settle="usdt"), self.exchange_info, "futures"
            )
            perps = {}
            for settle in self.PERP_SETTLE:
                perp = self.parser.parse_tickers(
                    await self.exchange._get_perp_tickers(settle), self.exchange_info, "perp"
                )
                perps.update(perp)
            return {**spot, **futures, **perps}

    async def get_klines(
        self, instrument_id: str, interval: str, start: int = None, end: int = None, num: int = None
    ) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in exchange_info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)
        _interval = self.parser.get_interval(interval)
        limit_map = {"spot": 1000, "futures": 2000, "perp": 2000}

        symbol_map = {
            "spot": "id",
            "futures": "name",
            "perp": "name",
        }

        method_map = {
            "spot": self.exchange._get_spot_klines,
            "futures": self.exchange._get_futures_klines,
            "perp": self.exchange._get_perp_klines,
        }

        params = {
            "symbol": info["raw_data"][symbol_map[market_type]],
            "interval": _interval,
            "limit": limit_map[market_type],
        }

        # handle special market type params
        if market_type in ["futures", "perp"]:
            params["settle"] = info["settle"].lower()

        results = {}
        query_end = None
        if start and end:
            query_end = str(end)[:10]
            while True:
                params.update({"end": query_end})
                result = self.parser.parse_klines(await method_map[market_type](**params), market_type, info)
                results.update(result)

                if len(result) < limit_map[market_type]:
                    break

                query_end = sorted(list(result.keys()))[0]

                if query_end < start:
                    break

                query_end = str(query_end)[:10]
                continue

            return {k: v for k, v in results.items() if start <= k <= end}

        elif num:
            while True:
                params.update({"end": query_end} if query_end else {})
                result = self.parser.parse_klines(await method_map[market_type](**params), market_type, info)
                results.update(result)

                if len(result) < limit_map[market_type] or len(results) >= num:
                    break
                query_end = str(sorted(list(result.keys()))[0])[:10]

            return sort_dict(results, ascending=True, num=num)

        else:
            raise ValueError("(start, end) or num must be provided")



from .base import BaseClient


class BitgetUnified(BaseClient):
    BASE_URL = "https://api.bitget.com"

    def __init__(self) -> None:
        super().__init__()
        self.base_endpoint = self.BASE_URL

    async def _get_spot_exchange_info(self):
        return await self._get(self.base_endpoint + "/api/v2/spot/public/symbols")

    async def _get_derivative_exchange_info(self, product_type: str):
        params = {"productType": product_type}
        return await self._get(self.base_endpoint + "/api/v2/mix/market/contracts", params=params)

    async def _get_spot_tickers(self):
        return await self._get(self.base_endpoint + "/api/v2/spot/market/tickers")

    async def _get_derivative_tickers(self, product_type: str):
        params = {"productType": product_type}
        return await self._get(self.base_endpoint + "/api/v2/mix/market/tickers", params=params)

    async def _get_spot_candlesticks(
        self, symbol: str, granularity: str, startTime: str = None, endTime: str = None, limit: str = None
    ):
        params = {"symbol": symbol, "granularity": granularity}
        if startTime:
            params["startTime"] = startTime
        if endTime:
            params["endTime"] = endTime
        if limit:
            params["limit"] = limit
        return await self._get(self.base_endpoint + "/api/v2/spot/market/candles", params=params)

    async def _get_derivative_candlesticks(
        self,
        symbol: str,
        productType: str,
        granularity: str,
        startTime: str = None,
        endTime: str = None,
        kLineType: str = "MARKET",
        limit: str = None,
    ):
        params = {"symbol": symbol, "productType": productType, "granularity": granularity, "kLineType": kLineType}
        if startTime:
            params["startTime"] = startTime
        if endTime:
            params["endTime"] = endTime
        if limit:
            params["limit"] = limit
        return await self._get(self.base_endpoint + "/api/v2/mix/market/candles", params=params)



import asyncio

from .exchanges.kucoin import KucoinFutures, KucoinSpot
from .parsers.kucoin import KucoinParser
from .utils import query_dict, sort_dict


class Kucoin(object):
    name = "kucoin"

    def __init__(self):
        self.spot = KucoinSpot()
        self.futures = KucoinFutures()
        self.parser = KucoinParser()

        self.exchange_info = {}

    @classmethod
    async def create(cls):
        instance = cls()
        instance.exchange_info = await instance.get_exchange_info()
        return instance

    async def close(self):
        await self.spot.close()
        await self.futures.close()

    async def get_exchange_info(self) -> dict:
        spot = self.parser.parse_exchange_info(
            await self.spot._get_symbol_list(), self.parser.spot_exchange_info_parser
        )
        futures = self.parser.parse_exchange_info(
            await self.futures._get_symbol_list(), self.parser.futures_exchange_info_parser
        )

        return {**spot, **futures}

    async def get_tickers(self, market_type: str = None) -> dict:
        async def _get_derivative_tickers():
            ids = list(query_dict(self.exchange_info, "is_futures == True or is_perp == True").keys())
            num_batch = 20
            results = {}
            for i in range(0, len(ids), num_batch):
                tasks = []
                for instrument_id in ids[i : i + num_batch]:
                    _symbol = self.exchange_info[instrument_id]["raw_data"]["symbol"]
                    tasks.append(self.futures._get_contract_info(_symbol))
                raw_tickers = await asyncio.gather(*tasks)
                parsed_tickers = self.parser.parse_derivative_tickers(raw_tickers, self.exchange_info)
                results.update(parsed_tickers)
            return results

        if market_type == "spot":
            return self.parser.parse_spot_tickers(await self.spot._get_tickers(), self.exchange_info)

        elif market_type == "futures":
            return await _get_derivative_tickers()

        else:
            spot_tickers = self.parser.parse_spot_tickers(await self.spot._get_tickers(), self.exchange_info)
            derivative_tickers = await _get_derivative_tickers()
            return {**spot_tickers, **derivative_tickers}

    async def get_klines(self, instrument_id: str, interval: str, start: int = None, end: int = None, num: int = None):
        info = self.exchange_info[instrument_id]
        market_type = "spot" if info["is_spot"] else "derivative"
        _symbol = info["raw_data"]["symbol"]
        _interval = self.parser.get_interval(interval, market_type)
        method_map = {
            "spot": self.spot._get_klines,
            "derivative": self.futures._get_klines,
        }
        limit_map = {
            "spot": 100,
            "derivative": 200,
        }

        params = {
            "symbol": _symbol,
            "granularity" if market_type == "derivative" else "type": _interval,
        }

        results = {}
        query_end = None
        if start and end:
            query_end = self.parser.parse_kucoin_timestamp(end, market_type)
            while True:
                params.update({"end": query_end})
                klines = self.parser.parse_klines(await method_map[market_type](**params), info, market_type)

                results.update(klines)
                if len(klines) < limit_map[market_type]:
                    break

                query_end = sorted(list(results.keys()))[0]
                if query_end < start:
                    break

                query_end = self.parser.parse_kucoin_timestamp(sorted(list(results.keys()))[0], market_type)
                continue
            return {k: v for k, v in results.items() if start <= k <= end}

        elif num:
            while True:
                params.update({"end": query_end} if query_end else {})
                klines = self.parser.parse_klines(await method_map[market_type](**params), info, market_type)

                results.update(klines)
                if len(results) >= num or len(klines) < limit_map[market_type]:
                    break

                query_end = (
                    int(sorted(list(results.keys()))[0] / 1000)
                    if market_type == "spot"
                    else sorted(list(results.keys()))[0]
                )
                continue
            return sort_dict(results, True, num)

        else:
            raise ValueError("Invalid parameters. (start, end) or (end, num) or (num) must be provided.")



from .base import BaseClient


class BinanceSpot(BaseClient):
    BASE_ENDPOINT = "https://api{}.binance.com"
    name = "binance"

    def __init__(self, api_key: str, api_secret: str, api_version: int = 3):
        super().__init__()
        self.base_endpoint = self.BASE_ENDPOINT.format(api_version)

        self.auth_data = {
            "api_key": api_key,
            "api_secret": api_secret,
        }

    async def _get_exchange_info(self):
        return await self._get(self.base_endpoint + "/api/v3/exchangeInfo")

    async def _get_ticker(self, symbol: str):
        return await self._get(self.base_endpoint + "/api/v3/ticker/24hr", params={"symbol": symbol})

    async def _get_tickers(self):
        return await self._get(self.base_endpoint + "/api/v3/ticker/24hr")

    async def _get_klines(
        self,
        symbol: str,
        interval: str,
        startTime: int = None,
        endTime: int = None,
        limit: int = 500,
        timeZone: str = "0",
    ):
        params = {"symbol": symbol, "interval": interval, "limit": limit, "timeZone": timeZone}
        if startTime:
            params["startTime"] = startTime
        if endTime:
            params["endTime"] = endTime
        return await self._get(self.base_endpoint + "/api/v3/klines", params=params)

    # Private endpoint
    async def _get_account_info(self):
        return await self._get(self.base_endpoint + "/api/v3/account", auth_data=self.auth_data)


class BinanceLinear(BaseClient):
    BASE_ENDPOINT = "https://fapi.binance.com"

    def __init__(self) -> None:
        super().__init__()
        self.linear_base_endpoint = self.BASE_ENDPOINT

    async def _get_exchange_info(self):
        return await self._get(self.linear_base_endpoint + "/fapi/v1/exchangeInfo")

    async def _get_ticker(self, symbol: str):
        return await self._get(self.linear_base_endpoint + "/fapi/v1/ticker/24hr", params={"symbol": symbol})

    async def _get_tickers(self):
        return await self._get(self.linear_base_endpoint + "/fapi/v1/ticker/24hr")

    async def _get_klines(
        self, symbol: str, interval: str, startTime: int = None, endTime: int = None, limit: int = 500
    ):
        params = {"symbol": symbol, "interval": interval, "limit": limit}
        if startTime:
            params["startTime"] = startTime
        if endTime:
            params["endTime"] = endTime
        return await self._get(self.linear_base_endpoint + "/fapi/v1/klines", params=params)


class BinanceInverse(BaseClient):
    BASE_ENDPOINT = "https://dapi.binance.com"

    def __init__(self) -> None:
        super().__init__()
        self.inverse_base_endpoint = self.BASE_ENDPOINT

    async def _get_exchange_info(self):
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/exchangeInfo")

    async def _get_ticker(self, symbol: str):
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/ticker/24hr", params={"symbol": symbol})

    async def _get_tickers(self):
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/ticker/24hr")

    async def _get_klines(
        self, symbol: str, interval: str, startTime: int = None, endTime: int = None, limit: int = 500
    ):
        params = {"symbol": symbol, "interval": interval, "limit": limit}
        if startTime:
            params["startTime"] = startTime
        if endTime:
            params["endTime"] = endTime
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/klines", params=params)



from .base import BaseClient


class GateioClient(BaseClient):
    BASE_URL = "https://api.gateio.ws/api/v4"

    def __init__(self) -> None:
        super().__init__()
        self.base_url = self.BASE_URL

    async def _get_currency_pairs(self):
        return await self._get(self.base_url + "/spot/currency_pairs")

    async def _get_perp_info(self, settle: str):
        return await self._get(self.base_url + f"/futures/{settle}/contracts")

    async def _get_futures_info(self, settle: str = "usdt"):
        return await self._get(self.base_url + f"/delivery/{settle}/contracts")

    async def _get_spot_tickers(self, currency_pair: str = None, timezone: str = None):
        params = {}
        if currency_pair:
            params["currency_pair"] = currency_pair
        if timezone:
            params["timezone"] = timezone
        return await self._get(self.base_url + "/spot/tickers", params=params)

    async def _get_perp_tickers(self, settle: str):
        return await self._get(self.base_url + f"/futures/{settle}/tickers")

    async def _get_futures_tickers(self, settle: str = "usdt"):
        return await self._get(self.base_url + f"/delivery/{settle}/tickers")

    async def _get_spot_klines(self, symbol: str, interval: str, start: int = None, end: int = None, limit: int = None):
        params = {"currency_pair": symbol, "interval": interval}
        if start:
            params["from"] = start
        if end:
            params["to"] = end
        if limit:
            params["limit"] = limit

        return await self._get(self.base_url + "/spot/candlesticks", params=params)

    async def _get_perp_klines(
        self, symbol: str, settle: str, interval: str, start: int = None, end: int = None, limit: int = None
    ):
        params = {"contract": symbol, "interval": interval}
        if start:
            params["from"] = start
        if end:
            params["to"] = end
        if limit:
            params["limit"] = limit

        return await self._get(self.base_url + f"/futures/{settle}/candlesticks", params=params)

    async def _get_futures_klines(
        self, symbol: str, settle: str, interval: str, start: int = None, end: int = None, limit: int = None
    ):
        params = {"contract": symbol, "interval": interval}
        if start:
            params["from"] = start
        if end:
            params["to"] = end

        if not start and not end and limit:
            params["limit"] = limit

        return await self._get(self.base_url + f"/delivery/{settle}/candlesticks", params=params)



from ..utils import query_dict
from .base import Parser


class HtxParser(Parser):

    response_keys = ["data", "ticks"]

    INTERVAL_MAP = {
        "spot": {
            "1m": "1min",
            "5m": "5min",
            "15m": "15min",
            "30m": "30min",
            "1h": "60min",
            "4h": "4hour",
            "1d": "1day",
            "1M": "1mon",
            "1w": "1week",
            "1y": "1year",
        },
        "linear": {
            "1m": "1min",
            "5m": "5min",
            "15m": "15min",
            "30m": "30min",
            "1h": "60min",
            "4h": "4hour",
            "1d": "1day",
            "1M": "1mon",
        },
        "inverse_perp": {
            "1m": "1min",
            "5m": "5min",
            "15m": "15min",
            "30m": "30min",
            "1h": "60min",
            "4h": "4hour",
            "1d": "1day",
            "1M": "1mon",
        },
        "inverse_futures": {
            "1m": "1min",
            "5m": "5min",
            "15m": "15min",
            "30m": "30min",
            "1h": "60min",
            "4h": "4hour",
            "1d": "1day",
            "1M": "1mon",
        },
    }

    def __init__(self):
        super().__init__()

    def check_htx_response(self, response: dict):
        if response["status"] != "ok":
            return {"code": 400, "status": "error", "data": response}
        else:
            for _key in self.response_keys:
                if _key in response:
                    return {"code": 200, "status": "success", "data": response[_key]}

    @property
    def spot_exchange_info_parser(self) -> dict:
        return {
            "active": (lambda x: x["state"] == "online"),
            "is_spot": True,
            "is_margin": False,
            "is_futures": False,
            "is_perp": False,
            "is_linear": True,
            "is_inverse": False,
            "symbol": (lambda x: self.parse_unified_symbol(x["bcdn"], x["qcdn"])),
            "base": (lambda x: self.parse_base_currency(x["bcdn"])),
            "quote": (lambda x: str(x["qcdn"])),
            "settle": (lambda x: str(x["qcdn"])),
            "multiplier": 1,  # spot and margin default multiplier is 1
            "leverage": (lambda x: float(x["lr"]) if x["lr"] else 1),  # spot and margin default leverage is 1
            "listing_time": (lambda x: int(x["toa"])),
            "expiration_time": None,  # spot not support this field
            "contract_size": 1,
            "tick_size": None,  # not yet implemented
            "min_order_size": None,  # not yet implemented
            "max_order_size": None,  # not yet implemented
            "raw_data": (lambda x: x),
        }

    @property
    def linear_exchange_info_parser(self) -> dict:
        return {
            "active": (lambda x: x["contract_status"] == 1),
            "is_spot": False,
            "is_margin": False,
            "is_futures": (lambda x: self.parse_is_futures(x["business_type"])),
            "is_perp": (lambda x: self.parse_is_perpetual(x["business_type"])),
            "is_linear": True,
            "is_inverse": False,
            "symbol": (lambda x: self.parse_unified_symbol(self.parse_pair(x)["base"], self.parse_pair(x)["quote"])),
            "base": (lambda x: self.parse_base_currency(self.parse_pair(x)["base"])),
            "quote": (lambda x: self.parse_pair(x)["quote"]),
            "settle": (lambda x: self.parse_pair(x)["quote"]),
            "multiplier": 1,
            "leverage": None,  # not yet implemented
            "listing_time": (lambda x: self.parse_str_to_timestamp(x["create_date"]) if x["create_date"] else None),
            "expiration_time": (
                lambda x: self.parse_str_to_timestamp(x["delivery_date"]) if x["delivery_date"] else None
            ),
            "contract_size": (lambda x: float(x["contract_size"])),
            "tick_size": (lambda x: float(x["price_tick"])),
            "min_order_size": None,  # not yet implemented
            "max_order_size": None,  # not yet implemented
            "raw_data": (lambda x: x),
        }

    @property
    def inverse_futures_exchange_info_parser(self):
        return {
            "active": (lambda x: x["contract_status"] == 1),
            "is_spot": False,
            "is_margin": False,
            "is_futures": True,
            "is_perp": False,
            "is_linear": False,
            "is_inverse": True,
            "symbol": (lambda x: self.parse_unified_symbol(x["symbol"], "USD")),
            "base": (lambda x: str(x["symbol"])),
            "quote": "USD",
            "settle": (lambda x: str(x["symbol"])),
            "multiplier": 1,
            "leverage": None,  # not yet implemented
            "listing_time": (lambda x: self.parse_str_to_timestamp(x["create_date"]) if x["create_date"] else None),
            "expiration_time": (lambda x: int(x["delivery_time"]) if x["delivery_time"] else None),
            "contract_size": (lambda x: float(x["contract_size"])),
            "tick_size": (lambda x: float(x["price_tick"])),
            "min_order_size": None,  # not yet implemented
            "max_order_size": None,  # not yet implemented
            "raw_data": (lambda x: x),
        }

    @property
    def inverse_perp_exchange_info_parser(self):
        return {
            "active": (lambda x: x["contract_status"] == 1),
            "is_spot": False,
            "is_margin": False,
            "is_futures": False,
            "is_perp": True,
            "is_linear": False,
            "is_inverse": True,
            "symbol": (lambda x: self.parse_unified_symbol(x["symbol"], "USD")),
            "base": (lambda x: str(x["symbol"])),
            "quote": "USD",
            "settle": (lambda x: str(x["symbol"])),
            "multiplier": 1,
            "leverage": None,  # not yet implemented
            "listing_time": (lambda x: self.parse_str_to_timestamp(x["create_date"]) if x["create_date"] else None),
            "expiration_time": None,
            "contract_size": (lambda x: float(x["contract_size"])),
            "tick_size": (lambda x: float(x["price_tick"])),
            "min_order_size": None,  # not yet implemented
            "max_order_size": None,  # not yet implemented
            "raw_data": (lambda x: x),
        }

    @staticmethod
    def parse_pair(response: dict) -> dict:
        if response["delivery_date"]:
            return {
                "base": response["pair"].split("-")[0],
                "quote": response["pair"].split("-")[1],
                "datetime": response["delivery_date"],
            }
        else:
            return {
                "base": response["pair"].split("-")[0],
                "quote": response["pair"].split("-")[1],
            }

    def parse_exchange_info(self, response: dict, parser: dict):
        response = self.check_htx_response(response)
        if response["code"] != 200:
            return response

        results = {}
        datas = response["data"]
        for data in datas:
            result = self.get_result_with_parser(data, parser)
            id = self.parse_unified_id(result)
            results[id] = result
        return results

    def get_htx_id_map(self, exchange_info: dict, market_type: str) -> dict:
        keys_map = {
            "spot": "sc",
            "linear": "contract_code",
            "inverse_perp": "contract_code",
            "inverse_futures": "symbol",
        }
        if market_type == "linear":
            infos = query_dict(exchange_info, "(is_perp == True or is_futures == True) and is_linear == True")
        elif market_type == "inverse_perp":
            infos = query_dict(exchange_info, "is_perp == True and is_inverse == True")
        elif market_type == "inverse_futures":
            infos = query_dict(exchange_info, "is_futures == True and is_inverse == True")
            return {self.parse_inverse_futures_symbol(v["raw_data"]): k for k, v in infos.items()}
        else:
            infos = self.query_dict(exchange_info, {f"is_{market_type}": True})
        return {v["raw_data"][keys_map[market_type]]: k for k, v in infos.items()}

    def parse_inverse_futures_symbol(self, datas: dict) -> str:
        contract_type_map = {
            "this_week": "CW",
            "next_week": "NW",
            "quarter": "CQ",
        }
        return f"{datas['symbol']}_{contract_type_map[datas['contract_type']]}"

    def parse_tickers(self, response: dict, exchange_infos: dict, market_type: str) -> dict:
        response = self.check_htx_response(response)
        if response["code"] != 200:
            return response

        method_map = {
            "spot": self.parse_spot_ticker,
            "linear": self.parse_linear_ticker,
            "inverse_perp": self.parse_inverse_perp_ticker,
            "inverse_futures": self.parse_inverse_futures_ticker,
        }

        keys_map = {
            "spot": "symbol",
            "linear": "contract_code",
            "inverse_perp": "contract_code",
            "inverse_futures": "symbol",
        }

        id_map = self.get_htx_id_map(exchange_infos, market_type)

        results = {}
        datas = response["data"]
        for data in datas:
            if data[keys_map[market_type]] not in id_map:
                print(f"Unmapped symbol: {data[keys_map[market_type]]} in {market_type}")
                continue
            instrument_id = id_map[data[keys_map[market_type]]]
            results[instrument_id] = method_map[market_type](data, exchange_infos[instrument_id])
        return results

    def parse_spot_ticker(self, response: dict, info: dict):
        return {
            "symbol": response["symbol"],
            "open_time": None,
            "close_time": None,
            "open": float(response["open"]),
            "high": float(response["high"]),
            "low": float(response["low"]),
            "last_price": float(response["close"]),
            "base_volume": float(response["amount"]),
            "quote_volume": float(response["vol"]),
            "price_change": float(response["close"]) - float(response["open"]),
            "price_change_percent": None,
            "raw_data": response,
        }

    def parse_linear_ticker(self, response: dict, info: dict):
        return {
            "symbol": response["contract_code"],
            "open_time": None,
            "close_time": int(response["ts"]),
            "open": float(response["open"]),
            "high": float(response["high"]),
            "low": float(response["low"]),
            "last_price": float(response["close"]),
            "base_volume": float(response["amount"]) * info["contract_size"],
            "quote_volume": float(response["vol"]),
            "price_change": float(response["close"]) - float(response["open"]),
            "price_change_percent": None,
            "raw_data": response,
        }

    def parse_inverse_perp_ticker(self, response: dict, info: dict):
        return {
            "symbol": response["contract_code"],
            "open_time": None,
            "close_time": int(response["ts"]),
            "open": float(response["open"]),
            "high": float(response["high"]),
            "low": float(response["low"]),
            "last_price": float(response["close"]),
            "base_volume": float(response["amount"]),
            "quote_volume": float(response["vol"]),
            "price_change": float(response["close"]) - float(response["open"]),
            "price_change_percent": None,
            "raw_data": response,
        }

    def parse_inverse_futures_ticker(self, response: dict, info: dict):
        return {
            "symbol": response["symbol"],
            "open_time": None,
            "close_time": int(response["ts"]),
            "open": float(response["open"]),
            "high": float(response["high"]),
            "low": float(response["low"]),
            "last_price": float(response["close"]),
            "base_volume": float(response["amount"]),
            "quote_volume": float(response["vol"]),
            "price_change": float(response["close"]) - float(response["open"]),
            "price_change_percent": None,
            "raw_data": response,
        }

    def get_market_type(self, info: dict) -> str:
        if info["is_spot"]:
            return "spot"
        elif info["is_linear"] and (info["is_futures"] or info["is_perp"]):
            return "linear"
        elif info["is_inverse"] and info["is_futures"]:
            return "inverse_futures"
        elif info["is_inverse"] and info["is_perp"]:
            return "inverse_perp"
        else:
            raise ValueError("Unknown market type")

    def parse_klines(self, response: dict, market_type: str, info: dict) -> dict:
        response = self.check_htx_response(response)
        if response["code"] != 200:
            return response

        method_map = {
            "spot": self.parse_spot_and_linear_kline,
            "linear": self.parse_spot_and_linear_kline,
            "inverse_perp": self.parse_inverse_perp_kline,
            "inverse_futures": self.parse_inverse_futures_kline,
        }

        results = {}
        datas = response["data"]
        for data in datas:
            timestamp = int(int(data["id"]) * 1000)
            results[timestamp] = method_map[market_type](data, info=info)
        return results

    def parse_spot_and_linear_kline(self, response: dict, info: dict) -> dict:
        return {
            "open": float(response["open"]),
            "high": float(response["high"]),
            "low": float(response["low"]),
            "close": float(response["close"]),
            "base_volume": float(response["amount"]),
            "quote_volume": float(response["vol"]),
            "close_time": None,
            "raw_data": response,
        }

    def parse_inverse_perp_kline(self, response: dict, info: dict) -> dict:
        return {
            "open": float(response["open"]),
            "high": float(response["high"]),
            "low": float(response["low"]),
            "close": float(response["close"]),
            "base_volume": float(response["amount"]),
            "quote_volume": float(response["vol"]) * info["contract_size"],
            "close_time": None,
            "raw_data": response,
        }

    def parse_inverse_futures_kline(self, response: dict, info: dict) -> dict:
        return {
            "open": float(response["open"]),
            "high": float(response["high"]),
            "low": float(response["low"]),
            "close": float(response["close"]),
            "base_volume": float(response["amount"]),
            "quote_volume": float(response["vol"]) * info["contract_size"],
            "close_time": None,
            "raw_data": response,
        }

    def get_interval(self, interval: str, market_type: str) -> str:
        if interval not in self.INTERVAL_MAP[market_type]:
            raise ValueError(
                f"Invalid interval: {interval} in {market_type}, must be one of {list(self.INTERVAL_MAP[market_type])}"
            )
        return self.INTERVAL_MAP[market_type][interval]



from .exchanges.bitget import BitgetUnified
from .parsers.bitget import BitgetParser
from .utils import query_dict, sort_dict


class Bitget(object):
    def __init__(self) -> None:
        self.bitget = BitgetUnified()
        self.parser = BitgetParser()
        self.exchange_info = {}

    async def close(self):
        await self.bitget.close()

    @classmethod
    async def create(cls):
        instance = cls()
        instance.exchange_info = await instance.get_exchange_info()
        return instance

    async def get_exchange_info(self, market_type: str = None):
        if market_type:
            if market_type == "spot":
                return self.parser.parse_exchange_info(
                    await self.bitget._get_spot_exchange_info(), self.parser.spot_exchange_info_parser
                )
            else:
                derivative = {}
                inverse = self.parser.parse_exchange_info(
                    await self.bitget._get_derivative_exchange_info("COIN-FUTURES"),
                    self.parser.derivative_exchange_info_parser,
                )
                derivative.update(inverse)

                for settle in self.parser.LINEAR_FUTURES_SETTLE:
                    sub_linear = self.parser.parse_exchange_info(
                        await self.bitget._get_derivative_exchange_info(f"{settle}-FUTURES"),
                        self.parser.derivative_exchange_info_parser,
                    )
                    derivative.update(sub_linear)

                instrument_id = list(query_dict(self.exchange_info, f"is_{market_type} == True").keys())
                return {k: v for k, v in derivative.items() if k in instrument_id}
        else:
            spot = self.parser.parse_exchange_info(
                await self.bitget._get_spot_exchange_info(), self.parser.spot_exchange_info_parser
            )

            inverse = self.parser.parse_exchange_info(
                await self.bitget._get_derivative_exchange_info("COIN-FUTURES"),
                self.parser.derivative_exchange_info_parser,
            )

            linear_futures = {}
            for settle in self.parser.LINEAR_FUTURES_SETTLE:
                sub_linear = self.parser.parse_exchange_info(
                    await self.bitget._get_derivative_exchange_info(f"{settle}-FUTURES"),
                    self.parser.derivative_exchange_info_parser,
                )
                linear_futures.update(sub_linear)

            return {**spot, **inverse, **linear_futures}

    async def get_tickers(self, market_type: str = None):
        if market_type:
            if market_type == "spot":
                return self.parser.parse_tickers(await self.bitget._get_spot_tickers(), self.exchange_info, "spot")
            else:
                derivative = {}
                inverse = self.parser.parse_tickers(
                    await self.bitget._get_derivative_tickers("COIN-FUTURES"), self.exchange_info, "derivative"
                )
                derivative.update(inverse)

                for settle in self.parser.LINEAR_FUTURES_SETTLE:
                    sub_linear = self.parser.parse_tickers(
                        await self.bitget._get_derivative_tickers(f"{settle}-FUTURES"), self.exchange_info, "derivative"
                    )
                    derivative.update(sub_linear)

                instrument_id = list(query_dict(self.exchange_info, f"is_{market_type} == True").keys())
                return {k: v for k, v in derivative.items() if k in instrument_id}
        else:
            spot = self.parser.parse_tickers(await self.bitget._get_spot_tickers(), self.exchange_info, "spot")
            inverse = self.parser.parse_tickers(
                await self.bitget._get_derivative_tickers("COIN-FUTURES"), self.exchange_info, "derivative"
            )
            linear_futures = {}
            for settle in self.parser.LINEAR_FUTURES_SETTLE:
                sub_linear = self.parser.parse_tickers(
                    await self.bitget._get_derivative_tickers(f"{settle}-FUTURES"), self.exchange_info, "derivative"
                )
                linear_futures.update(sub_linear)
            return {**spot, **inverse, **linear_futures}

    async def get_klines(
        self, instrument_id: str, interval: str, start: int = None, end: int = None, num: int = None
    ) -> dict:
        if instrument_id not in self.exchange_info:
            return {}

        market_type = self.parser.get_market_type(self.exchange_info[instrument_id])
        _symbol = self.exchange_info[instrument_id]["raw_data"]["symbol"]
        _interval = self.parser.get_interval(interval, market_type)
        limit = 300

        method_map = {
            "spot": self.bitget._get_spot_candlesticks,
            "derivative": self.bitget._get_derivative_candlesticks,
        }

        params = {
            "symbol": _symbol,
            "granularity": _interval,
            "limit": limit,
        }
        if market_type == "derivative":
            params.update({"productType": self.parser.get_product_type(self.exchange_info[instrument_id])})

        results = {}
        query_end = None
        if start and end:
            query_end = end
            while True:
                params.update({"endTime": query_end})

                result = self.parser.parse_klines(await method_map[market_type](**params), market_type)
                results.update(result)

                if not result or len(result) < limit:
                    break

                query_end = sorted(list(result.keys()))[0]

                if query_end < start:
                    break

            return {k: v for k, v in results.items() if start <= k <= end}
        elif num:
            while True:
                params.update({"endTime": query_end})

                result = self.parser.parse_klines(await method_map[market_type](**params), market_type)
                results.update(result)

                if not result or len(result) < limit:
                    break

                query_end = sorted(list(result.keys()))[0]

                if len(results) >= num:
                    break
            return sort_dict(results, ascending=True, num=num)

        else:
            return {"code": 400, "msg": "(start, end) or (num) must be provided at least one."}
